// This file is autogenerated. DO NOT EDIT
#include <robotpy_build.h>
{% if has_vcheck %}
#include <functional>
{% endif %}

{% for inc in extra_includes_first %}
#include <{{ inc }}>
{% endfor %}

#include <{{ rel_fname }}>

{% for inc in type_caster_includes %}
#include <{{ inc }}>
{% endfor %}

{% if need_operators_h %}
#include <pybind11/operators.h>
{% endif %}

{% for decl in using_declarations %}
using {{ decl.format() }};
{% endfor %}

{% for cls in classes_with_trampolines %}
#define RPYGEN_ENABLE_{{ cls.full_cpp_name_identifier }}_PROTECTED_CONSTRUCTORS
#include <rpygen/{{ cls.full_cpp_name_identifier }}.hpp>
{% endfor %}